function ggpm_demo(mymod)
% demo of all likelihood functions in GGPM toolbox
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% 2013-11-27 : ABC - update to GPML 3.4
%                    - used lik to directly calculate predictive distribution plot
%                    - cleaned up likelihoods, fixed numbering
%                    - added debugging (likelihood, derivative checks)
%                    - (removed own KLD inference)
%

%clear
if nargin<1
  mymod=[];
end

DEBUGMODE = 0;  % no debugging
% DEBUGMODE = 1;  % test consistency of the likelihood function
% DEBUGMODE = 2;  % same as above, but also dump results into a mat file


%%% prompt for model type %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

fprintf('*** GGPM Demo *******************************************\n');
fprintf('    LIKELIHOOD               OUTPUT\n');
fprintf('1)  Gaussian                 {reals}\n');
fprintf('10) Bernoulli (GPC)          {+1,-1}\n');
fprintf('15) Binomial                 {0/N,...,N/N}\n');
fprintf('20) Poisson                  {counting}\n');
fprintf('21) linearized Poisson       {counting}\n');
fprintf('22) COM-Poisson              {counting}\n');
fprintf('23) linearized COM-Poisson   {counting}\n');
fprintf('30) Geometric                {counting}\n');
fprintf('31) Negative Binomial        {counting}\n');
fprintf('40) Gamma (shape hyp)        {positive reals}\n');
fprintf('41) GammaScale (scale hyp)   {positive reals}\n');
fprintf('42) Chi-Squared              {positive reals}\n');
fprintf('43) Exponential              {positive reals}\n');
fprintf('44) InvGauss                 {positive reals}\n');
fprintf('50) Beta                     [0,1]\n');

if isempty(mymod)
  mymod = input('Select a likelihood model: ');
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% setup random numbers
myseed    = 200;

if (myseed > 0)
  randn('state', myseed);
  rand('state', myseed);
end

% defaults
covfunc  = {@covSEiso};
meanfunc = @meanZero;
hypinit = [];
maxscdf = 1;
Ydiscrete = 0;
models = {};

%%% select model and data %%%
switch(mymod)
    
  % In each case, the model and data generation method are defined.
  %
  % Variables are used to automatically generate toy data:
  %   feta(X)   -- latent function given X values
  %   fmu(eta)  -- mean function given latent values
  %   fy(mu)    -- sample Y given mean values
  %   fy2f(Y)   -- reverse mapping from Y to latent eta [for visualization] 
  %
  %   Yrange    -- range of Y values to compute log probabilities [for visualization]
  %   Ydiscrete -- 1 if Y is discrete, 0 if Y is continuous [for visualization]
  %
  % Variables defining the GP models:
  %   covfunc           -- covariance function (see GPML)
  %   models{i}.name    -- name of the method [for plotting]
  %   models{i}.likfunc -- likelihood function
  %   models{i}.inffunc -- inference function
  %

  
  %%% Gaussian %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 1
    feta   = @(X) 2*sinc(X/2)+0;
    fmu    = @(eta) eta;
    fy     = @(mu) mu + sqrt(0.1)*randn(size(mu));
    fy2f   = @(Y) Y;
    Yrange = linspace(-5,5,200);
    covfunc = {'covSEiso'};

    models{1}.name    = 'ExGauss-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExGauss};
    models{2}.name    = 'ExGauss-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExGauss};
    models{3}.name    = 'ExGauss-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExGauss};
    %models{4}.name    = 'ExGauss-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExGauss}; 
    
    
    %%% Bernoulli (equivalent to GPC) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 10
    feta   = @(X)   6*exp(-(1/8)*(X-2).^2) - 3;
    fmu    = @(eta) logistic(1, eta);
    fy     = @(mu)  binornd(1, mu);
    fy2f   = @(Y)   4*(Y-0.5);
    Yrange = [0, 1];
    Ydiscrete = 1;
    covfunc = {'covSEiso'};  

    models{1}.name    = 'ExBernoulli-Taylor';  
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExBernoulli};
    models{2}.name    = 'ExBernoulli-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExBernoulli};    
    models{3}.name    = 'ExBernoulli-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExBernoulli};    
    %models{4}.name    = 'ExBernoulli-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExBernoulli}; 
    

    %%% Binomial (Fixed N) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 15
    N = 10;
    feta   = @(X)   6*exp(-(1/8)*(X-2).^2) - 3;
    fmu    = @(eta) logistic(1, eta);
    fy     = @(mu)  binornd(N, mu)/N;
    fy2f   = @(Y)   4*(Y-0.5);
    hypinit.lik = log(N-1);
    Yrange = [0:N]/N;   
    Ydiscrete = 1;
    covfunc = {'covSEiso'};  

    models{1}.name    = 'ExBinomial-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExBinomialFixedN};
    models{2}.name    = 'ExBinomial-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExBinomialFixedN};
    models{3}.name    = 'ExBinomial-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExBinomialFixedN};    
    %models{4}.name    = 'ExBinomial-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExBinomialFixedN};    
    
    %%%  Poisson %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 20
    feta   = @(X)   0.2*(X+2);
    fmu    = @(eta) exp(eta);
    fy     = @(mu)  poissrnd(mu);
    fy2f   = @(Y)   log( Y.*(Y~=0) + exp(-1).*(Y==0) );
    Yrange = [0:50];
    Ydiscrete = 1;
    covfunc = {'covLINone'};

    models{1}.name    = 'ExPoisson-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExPoisson};
    models{2}.name    = 'ExPoisson-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExPoisson};
    models{3}.name    = 'ExPoisson-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExPoisson};
    %models{4}.name    = 'ExPoisson-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExPoisson};    
    %%% Linearized Poisson %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 21
    feta   = @(X)   4*X+1;
    fmu    = @(eta) max(eta,0);
    fy     = @(mu)  poissrnd(mu);
    fy2f = @(Y) Y;    
    Yrange = [0: 10 : 70];
    Ydiscrete = 1;
    covfunc = {'covSum', {'covLINone', 'covConst'}};
    
    models{1}.name    = 'ExPoissonLinear-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExPoissonLinear};
    models{2}.name    = 'ExPoissonLinear-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExPoissonLinear};
    models{3}.name    = 'ExPoissonLinear-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExPoissonLinear};
    %models{4}.name    = 'ExPoissonLinear-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExPoissonLinear};    
    %%% COM Poisson %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 22
    overdisp = 2.0
    feta   = @(X)   0.2*(X+2);
    fmu    = @(eta) exp(eta);
    fy     = @(mu) max(round(mu + overdisp*(poissrnd(mu)-mu)), 0);
    fy2f   = @(Y)   log( Y.*(Y~=0) + exp(-1).*(Y==0) );
    Yrange = [0:50];
    Ydiscrete = 1;
    covfunc = {'covLINone'};

    models{1}.name    = 'ExCOMPoisson-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExCOMPoisson};
    models{2}.name    = 'ExCOMPoisson-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExCOMPoisson};
    models{3}.name    = 'ExCOMPoisson-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExCOMPoisson};
    %models{4}.name    = 'ExCOMPoisson-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExCOMPoisson}; 
    
    % Linearized COM Poisson  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 23
    overdisp = 0.5;
    feta   = @(X)   4*X+1;
    fmu    = @(eta) max(eta,0);
    fy = @(mu) max(round(mu + overdisp*(poissrnd(mu)-mu)), 0);
    fy2f = @(Y) Y;    
    Yrange = [0:20:70];
    Ydiscrete = 1;
    covfunc = {'covSum', {'covLINone', 'covConst'}};
      
    % when overdisp=2, problems with w negative -- bad with long-tailed distributions?
    models{1}.name = 'ExCOMPoissonLinear-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExCOMPoissonLinear};
      
    % when overdisp=2, derivative problem? could be Z approximation.
    models{2}.name = 'ExCOMPoissonLinear-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExCOMPoissonLinear};
    
    models{3}.name = 'ExCOMPoissonLinear-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExCOMPoissonLinear};
    
    %models{4}.name    = 'ExCOMPoissonLinear-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExCOMPoissonLinear}; 
    
    %%% geometric (Fixed N) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 30
    feta   = @(X)   6*exp(-(1/8)*(X-2).^2) - 3;
    fmu    = @(eta) exp(eta);
    fy     = @(mu)  geornd(1./(1+mu));
    fy2f   = @(Y)   log( Y.*(Y~=0) + exp(-1).*(Y==0) );
    Yrange = [0:50];
    Ydiscrete = 1;
    covfunc = {'covSEiso'};  
    % covfunc = {'covSum', {'covNoise', 'covSEiso'}};

    models{1}.name    = 'ExGeometric-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExGeometric};
    models{2}.name    = 'ExGeometric-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExGeometric};
    models{3}.name    = 'ExGeometric-EP-NInt';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExGeometric};     
    %models{4}.name    = 'likExGeometric-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExGeometric}; 
    
    %%% Negative Binomial %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 31
    phi  = 0.3; %0.5;
    phi1 = 1/phi;
    feta = @(X) X*0.2+0;
    fmu  = @(eta) (1+exp(-eta)).^(-phi1)./(1-(1+exp(-eta)).^(-phi1))*phi1;
    fy   = @(mu) nbinrnd(phi1, 1./(1+mu*phi));
    fy2f = @(Y)   log( phi*Y.*(Y~=0) + phi*exp(-10).*(Y==0) );
    Yrange = linspace(0,50,200);
    covfunc = {'covSEiso'}; 
    hypinit.lik = log(2);
    
    models{1}.name    = 'ExNegBinomial-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExNegBinomial2};   
    
    models{2}.name    = 'ExNegBinomial-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExNegBinomial2};
    
    models{3}.name    = 'ExNegBinomial-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExNegBinomial2};
    
    %models{4}.name    = 'likExNegBinomial2-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExNegBinomial2};   
    
    %%% Gamma (shape hyperparameter) %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 40
    nu   = 2; %0.5;
    feta = @(X) X*0.2+0;
    fmu  = @(eta) exp(eta);
    fy   = @(mu) gamrnd(nu, mu/nu);
    fy2f = @(Y)   log( Y.*(Y~=0) + exp(-10).*(Y==0) );
    Yrange = linspace(0,50,200);
    covfunc   = {'covLINone'};  
    
    models{1}.name    = 'ExGamma-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExGamma};    
    models{2}.name    = 'ExGamma-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExGamma};
    models{3}.name    = 'ExGamma-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExGamma};
    %models{4}.name    = 'ExGamma-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExGamma}; 
        
    %%% Gamma (scale hyperparameter) %%%%%%%%%%%%%%%%%%%%%%% 
  case 41
    feta = @(X) X*0.2 + 0;
    fmu  = @(eta) exp(eta);
    myphi = 2;  %0.5
    fy   = @(mu) gamrnd(mu/myphi, myphi);
    fy2f = @(Y)   log( Y.*(Y~=0) + exp(-10).*(Y==0) );
    Yrange = linspace(0,50,100);
    covfunc = {'covLINone'};

    models{1}.name    = 'ExGammaScale-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExGammaScale};
    models{2}.name    = 'ExGammaScale-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExGammaScale};
    
    % TODO (EP w/ b(theta,phi) and T(y)) NaN problem
    models{3}.name    = 'ExGammaScale-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExGammaScale};
    
  %  models{4}.name    = 'ExGammaScale-KLD';
  %  models{4}.inffunc = @infKLD_diag;
  %  models{4}.likfunc = {@likExpo_generic, @likExGammaScale}; 
     
    %%% Chi-Squared Distribution
  case 42
    nu   = 2;
    feta = @(X) X*0.2+0;
    fmu  = @(eta) exp(eta);
    fy   = @(mu) chi2rnd(mu);   
    fy2f = @(Y)   log( Y.*(Y~=0) + exp(-10).*(Y==0) );
    Yrange = linspace(0,50,100);
    covfunc   = {'covLINone'}; 
    covfunc = {'covSEiso'}; 
    % covfunc = {'covSum', {'covLINone', 'covSEiso'}};
    
    models{1}.name    = 'ExGammaScale2-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExGammaScale2};
    
    models{2}.name    = 'ExGammaScale2-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExGammaScale2};
    
    % ?? derivative check is bad! but deriv @ convergence = 0 ?????
    models{3}.name    = 'ExGammaScale2-EP-NInt';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExGammaScale2};
    
  %  models{4}.name    = 'likExGammaScale2-KLD-diag';
  %  models{4}.inffunc = @infKLD_diag;
  %  models{4}.likfunc = {@likExpo_generic, @likExGammaScale2};    
    
    %%% Exponential %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 43
    feta = @(X) X*0.2+0;
    fmu  = @(eta) exp(eta);
    fy   = @(mu) exprnd(mu);
    fy2f = @(Y)   log( Y.*(Y~=0) + exp(-10).*(Y==0) );
    Yrange = linspace(0,80,200);
    covfunc   = {'covLINone'};  

    models{1}.name    = 'ExExpo-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExExponential};
    models{2}.name    = 'ExExpo-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExExponential};
    models{3}.name    = 'ExExpo-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExExponential};
  %  models{4}.name    = 'ExExpo-KLD';
  %  models{4}.inffunc = @infKLD_diag;
  %  models{4}.likfunc = {@likExpo_generic, @likExExponential};    
    
    %%% InvGauss %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  case 44
    %feta = @(X) X*0.2+2;
    %lambda = 15;
    %lambda = 0.3;
    
    feta = @(X) X*0.2+0;
    lambda = 10;
    %lambda = 0.3;
    fmu  = @(eta) exp(0.5*eta)/sqrt(2);
    fy   = @(mu) invgrnd(mu, lambda);
    fy2f = @(Y)   log(2) + 2*log(Y);
    Yrange = linspace(0,10,100);
    covfunc   = {'covSum', {'covLINone', 'covConst'}};  

    % not that good when large variance
    models{1}.name    = 'ExInvGauss-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExInvGauss};          
    
    models{2}.name    = 'ExInvGauss-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExInvGauss};
    
    % some numerical instability for small eta)
    models{3}.name    = 'ExInvGauss-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExInvGauss};
    
 %   models{4}.name    = 'ExInvGauss-KLD';
 %   models{4}.inffunc = @infKLD_diag;
 %   models{4}.likfunc = {@likExpo_generic, @likExInvGauss}; 
    
    %%% Beta %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % not so stable for shape<1
  case 50
    feta = @(X) X*0.2+0;
    shp  = 15;
    fmu  = @(eta) exp(eta - log(1+exp(eta)));
    fy   = @(mu) betarnd(mu*shp, (1-mu)*shp);
    fy2f = @(y)  log(y) - log(1-y);
    Yrange = linspace(0,1,100);
    covfunc = {'covLINone'};
    maxscdf = 5;

    % OK (for shp>2)  inconsistent derivatives for shp=1
    models{1}.name = 'ExBeta-Taylor';
    models{1}.inffunc = @infTaylor;
    models{1}.likfunc = {@likExpo_generic, @likExBeta};      

    models{2}.name = 'ExBeta-Laplace';
    models{2}.inffunc = @infLaplace;
    models{2}.likfunc = {@likExpo_generic, @likExBeta};
    
    models{3}.name = 'ExBeta-EP';
    models{3}.inffunc = @infEP;
    models{3}.likfunc = {@likExpo_generic, @likExBeta};
    
    %models{4}.name    = 'ExBeta-KLD';
    %models{4}.inffunc = @infKLD_diag;
    %models{4}.likfunc = {@likExpo_generic, @likExBeta}; 
    
  otherwise
    error('bad selection!');
    
end % end switch
  
  
%%% SAMPLE DATA %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ddim = 1;
numpoints = 50;
X = 5*randn(ddim,numpoints);   % each column is a data point
eta = feta(X);                 % compute the latent value given X
mu  = fmu(eta);                % compute the mean given eta

if nargin(fy) == 1
  Y   = fy(mu);                  % sample Y given mean
else
  Y   = fy(mu, X);
end
  
  
% range of X
dX = max(X)-min(X);
dY = max(real(Y))-min(real(Y));
  
% range of X for plotting
minX = min(X)-0.4*dX;
maxX = max(X)+0.4*dX;

% range of Y for plotting
minY = min(real(Y))-0.2*dY;
maxY = max(real(Y))+0.2*dY;

% values of X for prediction
Xall = linspace(minX, maxX, 500);

% true latent values and means for X
Fall = feta(Xall);
Mall = fmu(Fall);


%%% INITIALIZE HYPERPARAMETERS %%%%%%%%%%%%%%%%%%%%%%
nhyp    = eval(feval(covfunc{:}));
if ~exist('hypinit') || ~isfield(hypinit, 'cov')
  hypinit.cov = zeros(nhyp,1);
end


%%% for each model %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for mm = 1 :length(models) 
  
  fprintf('*** %s **************************************************\n', models{mm}.name);
  
  % check if imaginary (confidence levels)
  if ~isreal(Y)    
    if strcmp(func2str(models{mm}.likfunc), 'likExGaussYConf')
      YY = Y;
    else
      % convert to real if the likelihood can't handle confidence levels
      YY = real(Y);
    end
  end
  
  nlhyp = eval(feval(models{mm}.likfunc{:}));
  if (nlhyp>0) && ~isfield(hypinit, 'lik')
   hypinit.lik = zeros(nlhyp,1);
  end
  
  % Added by Lifeng Shang for checking output format Y
  likFun = models{mm}.likfunc{2};
  if isfield(hypinit, 'lik')
   likStruct = feval(likFun, hypinit.lik); % the second input is useless, 1 can be any natural number
  else
   likStruct = feval(likFun, 0); % the second input is useless, 1 can be any natural number
  end
  YY = likStruct.chky(Y);
  
  %%% DEBUGGING %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
  if DEBUGMODE 
    % check likelihood function
    models{mm}.liktest = lik_check(hypinit, models{mm}.likfunc, Y(:), models{mm}.inffunc);
        
    % check derivatives
    models{mm}.derivtest = deriv_check(hypinit, models{mm}.inffunc, meanfunc, covfunc, ...
      models{mm}.likfunc, X', Y(:));    
  
    % get initial result (at fixed hyp) -- for testing
    % nlZ and derivative
    [init_nlZ, init_dnlZ] = gp(hypinit, models{mm}.inffunc, meanfunc, covfunc, models{mm}.likfunc, X', YY(:));
    % prediction
    [init_ymu, init_ys2, init_fmu, init_fs2] = gp(hypinit, models{mm}.inffunc, meanfunc, covfunc, models{mm}.likfunc, X', YY(:), Xall');
    % calculate log p(Y)
    Ytmp = repmat(Yrange(:)', [size(Xall,2) 1]);
    if isfield(hypinit, 'lik')
      hyplik = hypinit.lik;
    else
      hyplik = [];
    end
    [init_lp] = feval(models{mm}.likfunc{:}, hyplik, Ytmp, init_fmu(:), init_fs2(:));
    % save to model
    models{mm}.init.hyp  = hypinit;
    models{mm}.init.nlZ  = init_nlZ;
    models{mm}.init.dnlZ = init_dnlZ;
    models{mm}.init.Ymu  = init_ymu;
    models{mm}.init.Ys2  = init_ys2;
    models{mm}.init.Fmu  = init_fmu;
    models{mm}.init.Fs2  = init_fs2;
    models{mm}.init.Ylp  = init_lp;
  end

  
  try
    %%% RUN MINIMIZATION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    hyp = minimize(hypinit, @gp, -500, models{mm}.inffunc, meanfunc, covfunc, models{mm}.likfunc, X', YY(:));
  
    % get final result
    [nlZ, dnlZ] = gp(hyp, models{mm}.inffunc, meanfunc, covfunc, models{mm}.likfunc, X', YY(:));
  
    fprintf('========\n');
    fprintf('nlZ = %g\n', nlZ);
    ff = fieldnames(hyp);
    for i=1:length(ff)
      fprintf('dnlZ.%s = %s\n', ff{i}, num2str(getfield(dnlZ, ff{i})'));
    end
    fprintf('========\n');
    
    % save some things
    models{mm}.hyp  = hyp;
    models{mm}.nlZ  = nlZ;
    models{mm}.dnlZ = dnlZ;
    
    
    %%% PREDICT %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    % calculate predited mean and variance
    [Ypred_mu Ypred_s2 Fpred_mu Fpred_s2] ...
      = gp(hyp, models{mm}.inffunc, meanfunc, covfunc, models{mm}.likfunc, ...
      X', YY(:), Xall');
    
    % use likelihood function to directly calculate log probabilities
    % (this is more efficient than using gp.m since likExpo_generic.m can
    % handle multiple Ys for the same input X_i.
    
    % all Y locations of interest
    Ytmp = repmat(Yrange(:)', [size(Xall,2) 1]);
    
    % calculate log p(Y)
    if isfield(hyp, 'lik')
      hyplik = hyp.lik;
    else
      hyplik = [];
    end
    [Ypred_lp] = feval(models{mm}.likfunc{:}, hyplik, Ytmp, Fpred_mu(:), Fpred_s2(:));
    
    % find the mode
    [foo, fooi] = max(Ypred_lp,[],2);
    pred.Ymode = Yrange(fooi);
    pred.Ymu   = Ypred_mu;
    pred.Ys2   = Ypred_s2;
    pred.Fmu   = Fpred_mu;
    pred.Fs2   = Fpred_s2;
    pred.Ylp   = Ypred_lp;
    
    models{mm}.pred     = pred;
    models{mm}.covfunc  = covfunc;
    models{mm}.meanfunc = meanfunc;
    
    %%% make plots %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    py = 2;
    px = length(models);
    
    figure(mymod)
    if (mm == 1)
      clf
    end
    amod = models{mm};
    
    %%% f %%%%%%%%%%%%%%%%%%%%%%
    subplot(py,px,mm+1*px)
    hold on
    
    errorbar_gray(Xall, amod.pred.Fmu, sqrt(amod.pred.Fs2), 1, [7 7 7]/8);
    hold on
    plot(Xall, Fall, 'r-');
    plot(X, fy2f(Y), 'r+', 'MarkerSize', 8, 'LineWidth', 2.0);
    
    plot(Xall, amod.pred.Fmu, 'b-');
    
    hold off
    grid
    ylabel('\eta(x)');
    xlabel('x');
    foo = sprintf('%g ', exp(amod.hyp.cov));
    title(['hyp.cov=' foo])
        
    %%% y %%%%%%%%%%%%%%%%%%%%%
    subplot(py,px,mm+0*px)
    hold on
    %plot(Xall, Mpred, 'b-');
    
    % set scaling
    tmpP  = exp(amod.pred.Ylp);
    %maxsc = min(max(tmpP(:)), 3*mean(tmpP(:)));
    maxsc = min(maxscdf, max(tmpP(:)));
    
    imagesc(Xall, Yrange, tmpP', [0 maxsc]);
    plot(Xall, Mall, 'r-');
    plot(X, real(Y), 'r+', 'MarkerSize', 8, 'LineWidth', 2.0);
    
    % plot empirical mean
    %empMean = sum(exp(amod.pred.Ylp).*repmat(Yrange,length(Xall),1),2);
    %if ~Ydiscrete
    %    empMean = empMean * (Yrange(2)-Yrange(1));
    %  end
    %  plot(Xall, empMean, 'm-');
    
    % mean, std, mode
    plot(Xall, amod.pred.Ymu, 'b-');
    plot(Xall, amod.pred.Ymu+sqrt(amod.pred.Ys2), 'c--');
    plot(Xall, amod.pred.Ymu-sqrt(amod.pred.Ys2), 'c--');
    plot(Xall, amod.pred.Ymode, 'k-');
    
    colormap pink
    %colorbar
    
    hold off
    axisx(minX, maxX);
    axisy(minY, maxY);
    grid
    
    set(gca, 'color', [0 0 1]);
    
    xlabel('x');
    ylabel('y');
    titstr = amod.name;
    if (isfield(amod.hyp, 'lik'))
      foo = sprintf('%g (%g)', exp(amod.hyp.lik), amod.hyp.lik);
      amod.name = [amod.name '; hyp.lik=' foo];
    end
    title(titstr);
    
  catch me
    fprintf('Error: could not complete minimization!\n');
    models{mm}.pred = [];
  end
end % end for models

%keyboard

% save if debugging (must be run in the GGPM root directory)
if (DEBUGMODE == 2)
  outname = sprintf('test_results/ggpm_demo_v1.0-4_mymod_%d.mat', mymod)
  save(outname);
end


