% 1D Example of using GGPM toolbox
%
% input X is 1-dimensional, which feeds into a mean function
% output Y is positive reals, from Gamma distribution
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

clear

% set the seed
myseed = 250;
if (myseed > 0)
  randn('state', myseed);
  rand('state', myseed);
end

echo on

% define the mean function
fmu = @(X) 20*exp(-(X-2).^2./12);

% sample input points X (each column is a data point), 
% calculate the mean, and sample Y
numpoints = 20;
nu        = 2;
X  = rand(1,numpoints)*10-5; % Uniform(-5,5)
mu = fmu(X);
Y  = gamrnd(nu, mu/nu);

% define grid points (in column order)
minX = -5; maxX = 5;
gridsp = 100;
Xall = linspace(minX, maxX, gridsp);

% calculate ground-truth mean function over grid
Muall = fmu(Xall);

% calculate ground-truth latent function over grid
Eta_all = log(Muall);

echo off

figure(1)
clf
subplot(2,2,1)
plot(Xall, Eta_all);
hold on
plot(X, log(Y), 'rx');
hold off
xlabel('x');
ylabel('\eta');
title('latent function');
legend({'latent function', 'log(y) training points'},'Location', 'SouthEast');
axis([-5 5 -2 4]);
grid on

subplot(2,2,2)
plot(Xall, Muall);
hold on
plot(X, Y, 'rx');
hold off
xlabel('x');
ylabel('y');
legend({'mean function', 'training points'},'Location', 'NorthWest');
title('mean function');
axis([-5 5 0 100]);
grid on

disp(['Plotting the latent function, mean function, training sample']);
disp('<press any key to continue>');
pause

echo on

% specify the covariance and mean functions
covfunc = {'covSEiso'};
meanfunc = @meanZero;

% use Taylor inference
inffunc = @infTaylor;

% specify the Gamma likelihood using the generic likelihood function
likfunc = {@likExpo_generic, @likExGamma};

% initialize the hyperparameters
nhyp = eval(feval(covfunc{:}));
inithyp.cov = zeros(nhyp,1);
nlik = eval(feval(likfunc{:}));
inithyp.lik = zeros(nlik,1);
  

% optimize the hyperparameters using maximum marginal likelihood (MLL).
% NOTE: GPML uses "row vector" format, so X and Y are transposed
hyp = minimize(inithyp, @gp, -200, inffunc, meanfunc, covfunc, likfunc, X', Y');

% the nu estimated using MLL (should be close to 2)
nu_hyp = exp(hyp.lik)

% calculate the predictive mean and variance for y and eta.
% NOTE: GPML uses "row vector" format, so X, Y, and Xall are transposed.
[Ypred_mu Ypred_s2 Epred_mu Epred_s2] = gp(hyp, inffunc, meanfunc, covfunc, likfunc, X', Y', Xall');

% calculate the uncertainty in Y, which is amount of overdispersion 
% versus a baseline Gamma distribution
unc = sqrt(Ypred_s2)./(Ypred_mu/sqrt(nu_hyp)) - 1;

echo off

subplot(2,2,3)
errorbar_gray(Xall, Epred_mu, sqrt(Epred_s2), 1, [7 7 7]/8);
hold on
plot(Xall, Epred_mu);
hold off
title('GGPM-Gamma latent mean');
axis([-5 5 -2 4]);
grid on

subplot(2,2,4)
errorbar_gray(Xall, Ypred_mu, sqrt(Ypred_s2), 1, [7 7 7]/8);
hold on
plot(Xall, Ypred_mu);
hold off
title('GGPM-Gamma function');
axis([-5 5 0 100]);
grid on

disp(['Plotting the latent mean and predictive function']);

return