% 2D Example of using GGPM toolbox
%
% input X is 2-dimensional, which feeds into a mean function
% output Y is positive reals, from Gamma distribution
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27


clear

% set the seed
myseed = 250;
if (myseed > 0)
  randn('state', myseed);
  rand('state', myseed);
end

echo on

% define the mean function
fmu = @(X) 20*exp(-sum(X.^2./repmat([12;6],[1,size(X,2)]),1));

% sample input points X (each column is a data point), 
% calculate the mean, and sample Y
numpoints = 20;
nu        = 2;
X  = rand(2,numpoints)*10-5; % Uniform(-5,5)
mu = fmu(X);
Y  = gamrnd(nu, mu/nu);

% define grid points (in column order)
minX = -5; maxX = 5;
gridsp = 50;
Xvals = linspace(minX, maxX, gridsp);
[Xgrid2 Xgrid1] = meshgrid(Xvals, Xvals);
Xall = [Xgrid1(:)'; Xgrid2(:)'];

% calculate ground-truth mean function over grid
Muall = fmu(Xall);
Muall_img = reshape(Muall, [gridsp, gridsp]);

% calculate ground-truth latent function over grid
Eta_all = log(Muall);
Eta_all_img = reshape(Eta_all, [gridsp gridsp]);

echo off

figure(1)
clf
subplot(2,2,1)
clims_eta = [-3 3];
imagesc(Xvals, Xvals, Eta_all_img, clims_eta);
colorbar
colormap gray
axis xy
title('latent function');

subplot(2,2,2)
surf(Xvals, Xvals, Muall_img);
title('mean function');

subplot(2,2,3)
clims = [0 30];
imagesc(Xvals, Xvals, Muall_img, clims);
hold on
plot(X(2,:), X(1,:), 'rx');
hold off
colorbar
colormap gray
axis xy
title('mean function w/ training points');

subplot(2,2,4)
[vorV, vorR] = voronoiDiagram(DelaunayTri(X'));
hold on
for i=1:length(vorR)
  tmp1 = vorR{i};
  tmp1(find(tmp1==1)) = [];  % remove first index
  patch(vorV(tmp1,2),vorV(tmp1,1),Y(i),'EdgeColor', 'blue');
end
plot(X(2,:), X(1,:), 'rx');
hold off
axisx(minX, maxX);
axisy(minX, maxX);
colorbar
title('sampled values for training points');

disp(['Plotting the latent function, mean function, training sample ' ...
      'points, and sample point values.']);
disp('<press any key to continue>');
pause

echo on

% specify the covariance and mean functions
covfunc = {'covSEiso'};
meanfunc = @meanZero;

% use Taylor inference
inffunc = @infTaylor;

% specify the Gamma likelihood using the generic likelihood function
likfunc = {@likExpo_generic, @likExGamma};

% initialize the hyperparameters
nhyp = eval(feval(covfunc{:}));
inithyp.cov = zeros(nhyp,1);
nlik = eval(feval(likfunc{:}));
inithyp.lik = zeros(nlik,1);
  

% optimize the hyperparameters using maximum marginal likelihood (MLL).
% NOTE: GPML uses "row vector" format, so X and Y are transposed
hyp = minimize(inithyp, @gp, -200, inffunc, meanfunc, covfunc, likfunc, X', Y');

% the nu estimated using MLL (should be close to 2)
nu_hyp = exp(hyp.lik)

% calculate the predictive mean and variance for y and eta.
% NOTE: GPML uses "row vector" format, so X, Y, and Xall are transposed.
[Ypred_mu Ypred_s2 Epred_mu Epred_s2] = gp(hyp, inffunc, meanfunc, covfunc, likfunc, X', Y', Xall');

% calculate the uncertainty in Y, which is amount of overdispersion 
% versus a baseline Gamma distribution
unc = sqrt(Ypred_s2)./(Ypred_mu/sqrt(nu_hyp)) - 1;

echo off

figure(2)
clf
subplot(2,2,1)
imagesc(Xvals, Xvals, reshape(Epred_mu, gridsp, gridsp), clims_eta);
colorbar
colormap gray
axis xy
title('GGPM latent mean');

subplot(2,2,2)
imagesc(Xvals, Xvals, reshape(sqrt(Epred_s2), gridsp, gridsp));
colorbar
colormap gray
axis xy
title('GGPM latent std');


subplot(2,2,3)
imagesc(Xvals, Xvals, reshape(Ypred_mu, gridsp, gridsp), clims);
colorbar
colormap gray
axis xy
title('GGPM y mean');

subplot(2,2,4)
imagesc(Xvals, Xvals, reshape(unc, gridsp, gridsp));
colorbar
axis xy
title('GGPM y uncertainty');

disp(['Plotting the latent mean and std functions, and the predictive ' ...
      'y mean and uncertainty.']);
disp(['Note how the uncertainty is higher in areas without training ' ...
      'data.,']);

return