function liktest = lik_check(hyp, likfunc, y, inffunc, randstd)
% check likelihood distribution and derivatives
%
%  liktest = lik_check(hyp, likfunc, Y, inffunc)
%
%        hyp = hyps
%    likfunc = likelihood function
%          Y = [n x 1]
%    inffunc = inference function
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27



% 2013-12-10: update to new GPML v3.4
%              - change infLaplace output for lik hyps
%


if isfield(hyp, 'lik')
  hlik = hyp.lik;
else
  hlik = [];
end

if nargin<5
  randstd = 1;
end

[n d] = size(y);

delta = 0.00001;

numchecks = 20;


% generic derivative checking structure
dfunc = {};  % [d, d_numerical]
dname = {};  % name of each derivative

if isa(inffunc, 'function_handle')
  inffuncstr = func2str(inffunc);
else
  inffuncstr = inffunc;
end

% check likelihood output
switch(inffuncstr)
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case {'infLaplace', 'infTaylor'}
  % Taylor basically uses the same derivatives
  
  dname = {'dlp', 'd2lp', 'd3lp', 'lp_dhyp', 'dlp_dhyp', 'd2lp_dhyp'};
  dfunc = cell(1,length(dname));
  
  for i=1:n
    eta = randstd*randn(1,1);
    
    % check logp derivatives wrt eta
    [lp, dlp, d2lp, d3lp] = feval(likfunc{:}, hlik, y(i), eta, 0, 'infLaplace');

    eta1 = eta+delta;
    eta2 = eta-delta;
    [lp1, dlp1, d2lp1, d3lp1] = feval(likfunc{:}, hlik, y(i), eta1, 0, 'infLaplace');
    [lp2, dlp2, d2lp2, d3lp2] = feval(likfunc{:},hlik, y(i), eta2, 0, 'infLaplace');
    
    dfunc{1}(i,1) = dlp;
    dfunc{1}(i,2) = (lp1-lp2) / (2*delta);
    dfunc{2}(i,1) = d2lp;
    dfunc{2}(i,2) = (dlp1-dlp2) / (2*delta);
    dfunc{3}(i,1) = d3lp;
    dfunc{3}(i,2) = (d2lp1-d2lp2) / (2*delta);    
    
    % check logp derivatives wrt hyp
    [lp_dhyp, dlp_dhyp, d2lp_dhyp] = feval(likfunc{:},hlik, y(i), eta, 0, 'infLaplace', 1);
    
    hlik1 = hlik+delta;
    hlik2 = hlik-delta;
    [lp1, dlp1, d2lp1, d3lp1] = feval(likfunc{:},hlik1, y(i), eta, 0, 'infLaplace');
    [lp2, dlp2, d2lp2, d3lp2] = feval(likfunc{:},hlik2, y(i), eta, 0, 'infLaplace');
   
    dfunc{4}(i,1) = lp_dhyp;
    dfunc{4}(i,2) = (lp1-lp2) / (2*delta);
    dfunc{5}(i,1) = dlp_dhyp;
    dfunc{5}(i,2) = (dlp1-dlp2) / (2*delta); 
    dfunc{6}(i,1) = d2lp_dhyp;
    dfunc{6}(i,2) = (d2lp1-d2lp2) / (2*delta); 
    
  end
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case 'infEP'
  dname = {'dlZ', 'd2lZ', 'dlZ_dhyp'};
  dfunc = cell(1,length(dname));
  
  eta = randstd*randn(n,1);
  s2  = (randstd*randn(n,1)).^2;
  
  % check lZ derivatives wrt eta
  [lZ, dlZ, d2lZ] = feval(likfunc{:},hlik, y, eta, s2, 'infEP');
  
  eta1 = eta+delta;
  eta2 = eta-delta;
  [lZ1, dlZ1, d2lZ1] = feval(likfunc{:},hlik, y, eta1, s2, 'infEP');
  [lZ2, dlZ2, d2lZ2] = feval(likfunc{:},hlik, y, eta2, s2, 'infEP');
  
  dfunc{1}(:,1) = dlZ;
  dfunc{1}(:,2) = (lZ1-lZ2) / (2*delta);
  dfunc{2}(:,1) = d2lZ;
  dfunc{2}(:,2) = (dlZ1-dlZ2) / (2*delta);
  
  % check lZ derivative wrt hyp
  [dlZhyp] = feval(likfunc{:},hlik, y, eta, s2, 'infEP', 1);
  hlik1 = hlik+delta;
  hlik2 = hlik-delta;
  
  [lZ1, dlZ1, d2lZ1] = feval(likfunc{:},hlik1, y, eta, s2, 'infEP');
  [lZ2, dlZ2, d2lZ2] = feval(likfunc{:},hlik2, y, eta, s2, 'infEP');
  
  dfunc{3}(:,1) = dlZhyp;
  dfunc{3}(:,2) = (lZ1-lZ2) / (2*delta);
  
  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
 case {'infKLD', 'infKLD_diag', 'infKLD_full', 'infKLD_diag_hyp', 'infKLD_full_hyp'}
  dname = {'df_dm', 'df_dv', 'df_dhyp'};
  dfunc = cell(1,length(dname));
  
  eta = randstd*randn(n,1);
  s2  = (randstd*randn(n,1)).^2;

  [fout] = feval(likfunc{:},hlik, y, eta, s2, 'infKLD');

  % check derivative wrt m (eta)
  eta1 = eta+delta;
  eta2 = eta-delta;
  [fout1] = feval(likfunc{:},hlik, y, eta1, s2, 'infKLD');
  [fout2] = feval(likfunc{:},hlik, y, eta2, s2, 'infKLD');
  dfunc{1}(:,1) = fout.df_dm;
  dfunc{1}(:,2) = (fout1.f-fout2.f) / (2*delta);
  
  % check derivative wrt v (s2)
  maxdelta = min(delta, s2);
  s21 = s2+maxdelta;
  s22 = s2-maxdelta;
  [fout1] = feval(likfunc{:},hlik, y, eta, s21, 'infKLD');
  [fout2] = feval(likfunc{:},hlik, y, eta, s22, 'infKLD');
  dfunc{2}(:,1) = fout.df_dv;
  dfunc{2}(:,2) = (fout1.f-fout2.f) ./ (2*maxdelta);
  
  % check derivative wrt hyp
  hlik1 = hlik+delta;
  hlik2 = hlik-delta;
  [fout1] = feval(likfunc{:},hlik1, y, eta, s2, 'infKLD');
  [fout2] = feval(likfunc{:},hlik2, y, eta, s2, 'infKLD');
  dfunc{3}(:,1) = fout.df_dhyp;
  dfunc{3}(:,2) = (fout1.f-fout2.f) / (2*delta);
  
 otherwise
  error('bad inffunc');
end


%%% derivative check output
for j=1:length(dname)
  fprintf('--- %s ---\n', dname{j});
  
  derror  = (dfunc{j}(:,2) - dfunc{j}(:,1));
  perror = abs((dfunc{j}(:,2) - dfunc{j}(:,1)) ./ dfunc{j}(:,1))*100;
  tmp = find( (dfunc{j}(:,1) == 0) .* (dfunc{j}(:,2) == 0) );
  perror(tmp) = 0;
  
  for i=1:size(dfunc{j},1)
    fprintf('  d=% 9.4g; dnum=% 9.4g; err=% 10.4g; %%=% 10.6f\n', ...
            dfunc{j}(i,1), dfunc{j}(i,2), derror(i), perror(i));
  end
  
  
  fprintf('  max%%=% 10.6f; mean%%=% 10.6f; med%%=% 10.6f; \n', ...
          max(perror), mean(perror), median(perror));

  perror_all{j} = perror;
end

% summary
fprintf('--- summary ---\n');
sumstr = '';
for j=1:length(dname)
  perror = perror_all{j};
  sumstr = [sumstr, ...
    sprintf('%15s: max%%=% 10.6f; mean%%=% 10.6f; med%%=% 10.6f; \n', ...
    dname{j}, max(perror), mean(perror), median(perror))];
  
end
disp(sumstr);

% output results
liktest.dfunc  = dfunc;
liktest.perror = perror_all;
liktest.sumstr = sumstr;