function lf = likExBeta(hyp)
% Beta likelihood function
%  support: 0<y<1 
%
%  hyp = log(shape)
%
%   theta = alpha / (alpha+beta) = mean
%   shape = alpha + beta
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% return number of likelihood parameters
if nargin<1, lf = 1; return; end

%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.chky    = @(y) checkY(y);

% get dispersion
shp    = exp(hyp(1));
logshp = hyp(1);

%%% setup Beta %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1/shp;
lf.da      = @() -1/shp;

lf.b       = @(theta) b_combo(theta, shp, 0);   % capture hyperparameter
lf.db      = @(theta) b_combo(theta, shp, 1);
lf.d2b     = @(theta) b_combo(theta, shp, 2);
lf.d3b     = @(theta) b_combo(theta, shp, 3);
lf.b_and_d = @(theta) b_and_d(theta, shp);

lf.lh      = @(y) logh(y, shp);
lf.dlh     = @(y) dlogh(y, shp);

lf.theta   = @(eta) exp(eta-log(1+exp(eta)));
lf.dtheta  = @(eta) exp(eta-2*log(1+exp(eta)));
lf.d2theta = @(eta) -1./(1+exp(eta)) ...
                      +3./((1+exp(eta)).^2) ...
                      -2./((1+exp(eta)).^3);
lf.d3theta = @(eta) 1./(1+exp(eta)) ...
    - 7 ./((1+exp(eta)).^2) ...
    + 12./((1+exp(eta)).^3) ...
    - 6 ./((1+exp(eta)).^4);

lf.fix_phi = 0;

%lf.T       = @(y) (log(y)-log(1-y));
lf.T       = @T;
lf.Tinv    = @(x) exp(x - log(1+exp(x)));

% special case: b has hyperparameter
lf.b_has_hyp  = 1;
lf.db_hyp     = @(theta) db_hyp(theta, shp);


%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% expansion point
lf.etahat = @(y) log(y)-log(1-y);


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function Ty = T(y)
Ty = log(y)-log(1-y);
% T(0) = T(1) = -inf ... i.e., not in support
ii = find((y==0) + (y==1));
if ~isempty(ii)
  Ty(ii) = -inf;
end

function [bout] = b_combo(theta, shp, d)
ts  = theta*shp;
sts = shp-ts;
switch(d)
 case 0
  bout = (gammaln(ts) + gammaln(sts))/shp;
 case 1
  bout = psi(0,ts) - psi(0,sts);
 case 2
  bout = shp*(psi(1,ts) + psi(1,sts));
 case 3
  bout = (shp^2)*(psi(2,ts) - psi(2,sts));
 otherwise
  error('bad d');
end

function [b, db, d2b, d3b] = b_and_d(theta, shp)
ts  = theta*shp;
sts = shp-ts;
if (nargout==2)
  tmp = psi(0, ts);
  psi0a = reshape(tmp, size(theta));
  tmp = psi(0, sts);
  psi0b = reshape(tmp, size(theta));
elseif (nargout==3)
  tmp = psi(0:1, ts);
  psi0a = reshape(tmp(1,:), size(theta));
  psi1a = reshape(tmp(2,:), size(theta));  
  tmp = psi(0:1, sts);
  psi0b = reshape(tmp(1,:), size(theta));
  psi1b = reshape(tmp(2,:), size(theta));
elseif (nargout==4)
  tmp = psi(0:2, ts);
  psi0a = reshape(tmp(1,:), size(theta));
  psi1a = reshape(tmp(2,:), size(theta));
  psi2a = reshape(tmp(3,:), size(theta));
  tmp = psi(0:2, sts);
  psi0b = reshape(tmp(1,:), size(theta));
  psi1b = reshape(tmp(2,:), size(theta));
  psi2b = reshape(tmp(3,:), size(theta));
end

b = (gammaln(ts) + gammaln(sts))/shp;
if nargout>1
  db = psi0a - psi0b;
  if (nargout>2)
    d2b = (psi1a + psi1b)*shp;
    if (nargout>3)
      d3b = (psi2a - psi2b)*shp*shp;
    end
  end
end

if 0
  % check against b_combo (OKAY)
  [Xb] = b_combo(theta, shp, 0);
  fprintf('b=%g; ', sum(abs(b-Xb)));
  if nargout>1
    [Xdb] = b_combo(theta, shp, 1);
    fprintf('db=%g; ', sum(abs(db-Xdb)));
    if nargout>2
      [Xd2b] = b_combo(theta, shp, 2);
      fprintf('d2b=%g; ', sum(abs(d2b-Xd2b)));
      if nargout>3
	[Xd3b] = b_combo(theta, shp, 3);
	fprintf('d3b=%g; ', sum(abs(d3b-Xd3b)));
      end
    end
  end
  fprintf('\n');
end
if 0
  % check derivatives (OKAY)
  delta = 0.0000001;
  if (nargout>1)
    Xb   = b_combo(theta+delta, shp, 0);
    Xdb  = b_combo(theta+delta, shp, 1);
    Xd2b = b_combo(theta+delta, shp, 2);
    Xd3b = b_combo(theta+delta, shp, 3);
  
    XXdb = (Xb - b)/delta;
    fprintf('db=%g; ', sum(abs((db-XXdb)./db)));
    if (nargout>2)
      XXd2b = (Xdb - db)/delta;
      fprintf('d2b=%g; ', sum(abs((d2b-XXd2b)./d2b)));
      if (nargout>3)
	XXd3b = (Xd2b - d2b)/delta;
	fprintf('d3b=%g; ', sum(abs((d3b-XXd3b)./d3b)));
      end
    end
    fprintf('\n');
  end
end


function [db, dbp, dbpp] = db_hyp(theta, shp)
ts  = theta*shp;
sts = shp-ts;
if (nargout==1)
  tmp = psi(0, ts);
  psi0a = reshape(tmp, size(theta));
  tmp = psi(0, sts);
  psi0b = reshape(tmp, size(theta));
elseif (nargout==2)
  tmp = psi(0:1, ts);
  psi0a = reshape(tmp(1,:), size(theta));
  psi1a = reshape(tmp(2,:), size(theta));  
  tmp = psi(0:1, sts);
  psi0b = reshape(tmp(1,:), size(theta));
  psi1b = reshape(tmp(2,:), size(theta));
elseif (nargout==3)
  tmp = psi(0:2, ts);
  psi0a = reshape(tmp(1,:), size(theta));
  psi1a = reshape(tmp(2,:), size(theta));
  psi2a = reshape(tmp(3,:), size(theta));
  tmp = psi(0:2, sts);
  psi0b = reshape(tmp(1,:), size(theta));
  psi1b = reshape(tmp(2,:), size(theta));
  psi2b = reshape(tmp(3,:), size(theta));
end

theta1 = 1-theta;

%b = (gammaln(ts) + gammaln(sts))/shp;
db = -(1/shp)*(gammaln(ts) + gammaln(sts)) + ...
     (1/shp)*(ts.*psi0a + sts.*psi0b);
if nargout>1
  %db = psi0a - psi0b;
  dbp = ts.*psi1a - sts.*psi1b;
  if (nargout>2)
    %d2b = (psi1a + psi1b)*shp;
    dbpp = shp*(psi1a+psi1b) + shp*(ts.*psi2a+sts.*psi2b);
  end
end

if 0
  % check derivatives  
  delta = 0.000001;
  X1b   = b_combo(theta, shp, 0);
  X1db  = b_combo(theta, shp, 1);
  X1d2b = b_combo(theta, shp, 2);
  X1d3b = b_combo(theta, shp, 3);
  X2b   = b_combo(theta, exp(log(shp)+delta), 0);
  X2db  = b_combo(theta, exp(log(shp)+delta), 1);
  X2d2b = b_combo(theta, exp(log(shp)+delta), 2);
  X2d3b = b_combo(theta, exp(log(shp)+delta), 3);
  
  XXdb = (X2b-X1b)/delta;
  fprintf('db=%g; ', sum(abs( (XXdb-db)./db)));
  if nargout>1
    XXdbp = (X2db-X1db)/delta;
    fprintf('dbp=%g; ', sum(abs( (XXdbp-dbp)./dbp)));
    if nargout>2
      XXdbpp = (X2d2b-X1d2b)/delta;
      fprintf('dbpp=%g; ', sum(abs( (XXdbpp-dbpp)./dbpp)));     
    end
  end
  fprintf('\n');
end


function lh = logh(y, shp)
lh = -log(y) + (shp-1)*log(1-y) + gammaln(shp);
% p(y=0) = p(y=1) = -inf
ii = find((y==0) + (y==1));
if ~isempty(ii)
  lh(ii) = -inf;
end


function dlh = dlogh(y, shp)
dlh = shp*(log(1-y) + psi(shp));



function y = checkY(y)
ee = 0;
ii = find(y<ee);
if ~isempty(ii)
  y(ii) = ee;
  warning('y has negative values!');
end
ii = find(y>(1-ee));
if ~isempty(ii)
  y(ii) = 1-ee;
  warning('y has >1 values!');
end

