function lf = likExCOMPoisson_internal(opt, hyp)
% Conway-Maxwell Poisson likelihood function (handles everything)
%   y in {0,1,2,3,...}
%
% we use the (mu, nu) parameterization in Guikema & Goffelt 2008.
%
%  hyperparameter:
%     nu = exp(hyp)
%    hyp = log(nu)
%
% internal options
%     opt.map = latent-to-mean mapping function: 'expo', 'linear'
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% 2010-11-05: updated S_terms to adaptively select MAXK
%             and uses Gaussian approximation for large means
% 2010-11-24: modified S_terms output format (more expandable)
%           : added outputs in db_hyp for better infTaylor
% 2010-11-25: Created likExCOMPoisson_internal to handle multiple versions
% 2010-11-25: BUGS: S_terms: fixed duplicate points in lbounds.
%                          : fixed iiextra indices (BUG when overdispersion)

% TODO: optimize
% TODO: test EP, Laplace
% TODO: better estimate of large-mean case: E[log n!]
% TODO: fix large-mean: EY = mu+1/(2nu) - 1/2
% TODO: fix precision problems when nu small (for linearCOM)

% return number of likelihood parameters
if nargin<2, lf = 1; return; end


%%% check y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.chky    = @(y) checkY(y);

%%% get hyperparameter %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nu = exp(hyp);

% bound nu (to prevent degeneration to bernoulli and geometric)
nu = min(nu,1e3);
nu = max(nu,1e-3);

%%% setup COM Poisson %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1/nu;
lf.da      = @() -1/nu;
%lf.b       = @(theta) b(theta, nu);   % capture hyperparameter
%lf.db      = @(theta) db(theta, nu);
%lf.d2b     = @(theta) d2b(theta, nu);
%lf.d3b     = @(theta) d3b(theta, nu);
lf.b       = @(theta) b_combo(theta, nu, 0);   % capture hyperparameter
lf.db      = @(theta) b_combo(theta, nu, 1);
lf.d2b     = @(theta) b_combo(theta, nu, 2);
lf.d3b     = @(theta) b_combo(theta, nu, 3);
lf.b_and_d = @(theta) b_and_d(theta, nu);
lf.lh      = @(y) -nu*logfactlut(y);
lf.dlh     = @(y) -nu*logfactlut(y);
lf.fix_phi = 0;  

switch(opt.map)
 case 'expo'
  lf.theta   = @(eta) eta;
  lf.dtheta  = @(eta) ones(size(eta));
  lf.d2theta = @(eta) zeros(size(eta));
  lf.d3theta = @(eta) zeros(size(eta));

  c = 0.001;
  % mu = exp(theta)
  % --> theta = log(mu)
  lf.etahat  = @(y) log(y+c);  % Taylor approximation
  
  % Ey = e^theta + 1/(2nu) - 1/2 (approximately)
  % --> y-0.5/nu+0.5 = e^theta
  % --> log(y-0.5/nu+0.5) = theta
  %lf.etahat = @(y) log(max(y-0.5/nu+0.5, c));
  
 case 'linear'
  lf.theta   = @(eta) theta(eta);
  lf.dtheta  = @dtheta;
  lf.d2theta = @d2theta;
  lf.d3theta = @d3theta;
  
  % theta(eta) = log(log(1+exp(eta)))
  % mu = exp(theta(eta)) = log(1+exp(eta))
  % --> y = log(1+exp(eta))
  %     e^y = 1+exp(eta)
  %     log(e^y-1) = eta
  
  c = 0.001;
  lf.etahat  = @(y) log(exp(y)-1+c);    % Taylor approximation
  
 otherwise
  error('bad opt.map');
end

% special case: b has hyperparameter
lf.b_has_hyp  = 1;
lf.db_hyp     = @(theta) db_hyp(theta, nu);

%%% SPECIAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% d = 0 --> b
% d = 1 --> db
% d = 2 --> d2b
function out = b_combo(theta, nu, d)
%if (size(theta,1) == 1) && (size(theta,2) > 1)
%  warning('theta not column!');
%end

% work on each column of theta
out = zeros(size(theta));
for i=1:size(theta,2)
  % do not recompute if theta column does not change
  if (i>1) && all(theta(:,i-1)==theta(:,i))
    out(:,i) = out(:,i-1);
    
  else
    switch(d)
     case 0    % b
      [logS] = S_terms(theta(:,i), nu, {'logS'});
      out(:,i) = logS / nu;
      
     case 1   % db
      [En] = S_terms(theta(:,i), nu, {'En'});
      out(:,i) = En;
      
     case 2   % d2b
      [Vn] = S_terms(theta(:,i), nu, {'Vn'});
      out(:,i) = nu*Vn;
      
     otherwise
      error('not implemented or invalid');
    end
    
    % check with old version
    %olderr = sum(abs(out(:,i) - b_combo_OLD(theta(:,i), nu, d)));
    %if (olderr > 1e-6)
    %  fprintf('olderr=%g; ', olderr);
    %end
    
  end
end

%%% consolidated b function %%%%%%%%%%%%%%%%%%
function [b, db, d2b, d3b] = b_and_d(theta, nu)
%if (size(theta,1) == 1) && (size(theta,2) > 1)
%  warning('theta not column!');
%end

b = zeros(size(theta));
if nargout > 1
  db = zeros(size(theta));
  if nargout > 2
    d2b = zeros(size(theta));
    if nargout > 3
      d3b = zeros(size(theta));
    end
  end
end

% work on each column of theta
for i=1:size(theta,2)
  % do not recompute if theta column does not change
  % this happens when computing lp for multiple y values
  if (i>1) && all(theta(:,i-1)==theta(:,i))    
    b(:,i) = b(:,i-1);
    if nargout > 1
      db(:,i) = db(:,i-1);
      if nargout > 2
	d2b(:,i) = d2b(:,i-1);
	if nargout > 3
	  d3b(:,i) = d3b(:,i-1);
	end      
      end
    end
  else

    if nargout>3
      [logS En Vn M3] = S_terms(theta(:,i), nu, {'logS', 'En', 'Vn', 'M3'});
    elseif nargout>2
      [logS En Vn] = S_terms(theta(:,i), nu, {'logS', 'En', 'Vn'});
    elseif nargout>1
      [logS En] = S_terms(theta(:,i), nu, {'logS', 'En'});
    else
      [logS] = S_terms(theta(:,i), nu, {'logS'});
    end
    b(:,i) = logS / nu;
    if nargout>1
      db(:,i) = En;
      if (nargout>2)
	d2b(:,i) = nu*Vn;
	if nargout>3
	  d3b(:,i) = (nu^2) * M3;
	end
      end
    end
  end
end

      

% NEWER VERSION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% mode is a cell array containing names of moments:
%   'logS'   -- log S
%   'En'     -- mean = E[n]
%   'Vn'     -- variance = E[n^2] - (E[n])^2
%   'M3'     -- 3rd moment = E[ (n-E[n])^3 ]
%   'Elnf'   -- E[log n!]
%   'Elnf2'  -- E[(log n!)^2]
%   'Enlnf'  -- E[n log n!]
%   'En2lnf' -- E[n^2 log n!]
% the output is ordered based on the ordering of mode.
function [varargout] = S_terms(theta, nu, mode)
%if (size(theta,2) ~= 1)
%  error('theta is not a column vector!');
%end

do_logS  = 0; do_En  = 0; do_Vn  = 0; do_M3 = 0;
do_Elnf  = 0; do_Elnf2 = 0; do_Enlnf = 0;
do_En2lnf = 0;
doNorm = 0;

% process input fields
for i=1:length(mode)
  switch(mode{i})
   case 'logS',   do_logS  = 1;    
   case 'En',     do_En    = 1; doNorm = 1;
   case 'Vn',     do_Vn    = 1; doNorm = 1;
   case 'M3',     do_M3    = 1; doNorm = 1;
   case 'Elnf',   do_Elnf  = 1; doNorm = 1;
   case 'Elnf2',  do_Elnf2 = 1; doNorm = 1;
   case 'Enlnf',  do_Enlnf = 1; doNorm = 1;
   case 'En2lnf', do_En2lnf = 1; doNorm = 1;
   otherwise
    error(sprintf('bad S_terms mode: %s', mode{i}));
  end
end

% variance and third moment require the mean
if (do_Vn || do_M3)
  do_En = 1;
end

% initialize storage
if do_logS,   logS   = zeros(size(theta)); end
if do_En,     En     = zeros(size(theta)); end
if do_Vn,     Vn     = zeros(size(theta)); end
if do_M3,     M3     = zeros(size(theta)); end
if do_Elnf,   Elnf   = zeros(size(theta)); end
if do_Elnf2,  Elnf2  = zeros(size(theta)); end
if do_Enlnf,  Enlnf  = zeros(size(theta)); end
if do_En2lnf, En2lnf = zeros(size(theta)); end


% chunks for running approximation
% 0-50    --> MAXK = 100
% 50-100  --> MAXK = 250
% 100-250 --> MAXK = 500
% 250-500 --> MAXK = 1000
% 500+    --> Gaussian approximation
bbounds  = [0 50 100 250 500 1000];
lbbounds = log(bbounds);
numbounds = 4;    % up to 250-500
iiextra = [];

DEBUGSUM = 0;
DEBUGPLOT = 0;
if DEBUGPLOT
  figure(999)
  clf
end
if DEBUGSUM
  debug_updates = zeros(size(theta));
end

for i=1:numbounds
  % find theta within this range
  ii = find( (lbbounds(i) < theta) .* (theta <= lbbounds(i+1)) );
  
  % get previous cases pushed to bigger range
  ii = [ii;iiextra];
  
  if DEBUGSUM
    debug_updates(ii) = debug_updates(ii)+1;
    if DEBUGSUM > 1
      fprintf('b%d=%d ', i, length(ii));
    end
  end
  
  if ~isempty(ii)
    MAXK = bbounds(i+2);
    
    % get log terms
    n = 0:MAXK;
    lts = nu*(theta(ii)*n - repmat(logfactlut(n),[length(ii),1]));

    % compute log(sum(exp(lts))) in a safe way
    % log(sum(exp(lts))) = log( sum(exp(lts-max)exp(max)) )
    %                      log( sum(exp(lts-max)) ) + max)
    ltsmax  = max(lts,[],2);
    mltsmax = repmat(ltsmax, [1,MAXK+1]);
    logsumts = log(sum(exp(lts-mltsmax),2)) + ltsmax;
  
    % compute bound
    lepsk = nu*theta(ii) - nu*log(MAXK+1);
    epsk = exp(lepsk);
    logbound = (MAXK+1)*nu*theta(ii) - nu*logfactlut(MAXK+1) ...
	- log(1-epsk) - logsumts;
    
    % error bound to high, so push these cases to the next round
    extra = find(logbound > -10);
    iiextra = ii(extra);    % BUG FIX: get actual indices
    if DEBUGSUM
      if ~isempty(iiextra)
	fprintf('MAXK=%d (nu=%g) bound=%g (pushing %d); ', MAXK, nu, ...
		max(logbound(:)), length(iiextra));
	if (DEBUGSUM < 2)
	  fprintf('\n');
	end
      end
    end
    
    if doNorm
      % create normalized distribution
      nts = exp(lts - repmat(logsumts, 1, MAXK+1));
    end
    
    % DEBUG
    if DEBUGPLOT
      figure(999)
      subplot(2,4,i)
      tmpii = setdiff((1:length(ii))', extra);
      if doNorm
	plot(nts(tmpii,:)')
      else
	plot(lts(tmpii,:)');
      end
      grid
      axisx(0,MAXK);
      title(sprintf('ii=%d [okay]', length(tmpii)));
      subplot(2,4,i+4)
      if  doNorm
	plot(nts(extra,:)');
      else
	plot(lts(extra,:)');
      end
      grid
      axisx(0,MAXK);
      title(sprintf('ii=%d [fail]', length(iiextra)));
      drawnow
      %if length(extra)>0
%	keyboard
 %     end
    end
    
    % get logS
    if do_logS
      logS(ii) = logsumts;    
    end
    
    % compute mean
    if do_En
      nn = repmat(n, [length(ii),1]);
      En(ii) = sum(nn .* nts, 2);
    end
    
    % compute variance
    if do_Vn
      nn2 = repmat(n.*n, [length(ii),1]);
      Vn(ii) = sum(nn2.*nts, 2) - En(ii).^2;
    end
    
    % compute third moment
    if do_M3
      %% compute third moment E[n^3] - (E[n])^3
      %nn3 = repmat(n.^3, [length(ii),1]);
      %M3(ii) = sum(nn3.*nts,2) - En(ii).^3;
      
      % compute third central moment E[ (n-En)^3 ]
      M3(ii) = sum( ((nn-repmat(En(ii),[1 length(n)])).^3).*nts, 2);
    end
    
    % compute E(log n!)
    if do_Elnf || do_Elnf2 || do_Enlnf || do_En2lnf
      lnf = repmat(logfactlut(n), [length(ii),1]);
      if do_Elnf
	Elnf(ii) = sum(lnf.*nts, 2);
      end
      if do_Elnf2
	Elnf2(ii) = sum(lnf.*lnf.*nts, 2);
      end
      if do_Enlnf
	Enlnf(ii) = sum(nn.*lnf.*nts, 2);
      end
      if do_En2lnf
	En2lnf(ii) = sum(nn2.*lnf.*nts, 2);
      end
    end
   
  end  % end if ~empty(ii)
end % end for i

% now handle cases with large mean...
ii = find(theta >= lbbounds(numbounds+1));


if DEBUGSUM > 1
  fprintf('b+=%d;\n', length(ii));
end

if ~isempty(ii)
  % for large mean, the distribution is approximately Gaussian
  % N(mu, mu/nu).
  % TODO: this method fails when theta(ii) is large --> mymu = inf
  %       usually this means that the data doesn't fit the exponential mean trend.
  mymu = exp(theta(ii));
  %mymu = exp(theta(ii)) + 1/(2*nu) - 0.5;
  if do_logS
    % match peaks of Gaussian and lts to get scaling
    % max(ts) / S = G(0,mu/nu)
    % logS = max(log(ts)) - G(0,mu/nu)
    mmu = floor(mymu);
    logS(ii) = nu*(mmu.*theta(ii) - gammaln(mmu)) + ...
	0.5*log(2*pi*mmu/nu);
  
    % stirling approximation when mmu large?
    % gammaln(mmu) = log(mmu!) 
    %              === mmu log(mmu/e) + 0.5 log(2 pi mmu)
    %                = mmu*(theta - 1) + 0.5 log(2pi) + 0.5 theta
    %
  end
  if do_En   % mean
    En(ii) = mymu;
  end
  if do_Vn   % variance
    Vn(ii) = mymu / nu;
  end
  if do_M3
    %% E[n^3] - E[n]^3 = 3mu var
    %M3(ii) = 3*mymu.*(mymu/nu);
    
    % E[ (n-En)^3 ] = 0;
    M3(ii) = 0;
  end
  if do_Elnf || do_Elnf2 || do_Enlnf || do_En2lnf   
    %warning('TODO! approximate now!');
    % just pick the mean and std points, and average
    tmps = sqrt(mymu/nu);
    tmp = [mymu, mymu-tmps, mymu+tmps];
    if do_Elnf
      Elnf(ii) = mean(gammaln(tmp), 2);
    end
    if do_Elnf2
      Elnf2(ii) = mean(gammaln(tmp).^2, 2);
    end
    if do_Enlnf
      Enlnf(ii) = mean(tmp.*gammaln(tmp), 2);
    end
    if do_En2lnf
      En2lnf(ii) = mean((tmp.^2).*gammaln(tmp), 2);
    end
  end    
%  keyboard
  if DEBUGSUM
    debug_updates(ii) = nan;
  end
end

if DEBUGSUM && 0
  fprintf('%d ', debug_updates);
  fprintf('\n');
  pause
end




% process outputs
for i=1:length(mode)
  switch(mode{i})
   case 'logS',   varargout{i} = logS;
   case 'En',     varargout{i} = En;
   case 'Vn',     varargout{i} = Vn;
   case 'M3',     varargout{i} = M3;
   case 'Elnf',   varargout{i} = Elnf;
   case 'Elnf2',  varargout{i} = Elnf2;
   case 'Enlnf',  varargout{i} = Enlnf;
   case 'En2lnf', varargout{i} = En2lnf;
   otherwise
    error(sprintf('bad S_terms mode: %s', mode{i}));
  end
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% derivative of b wrt hyperparameter 
function [db, dbp, dbpp] = db_hyp(theta, nu)

if nargout>2
  [logS, En, Elnf, Vn, Enlnf, M3, En2lnf] = S_terms(theta, nu, ...
     {'logS', 'En', 'Elnf', 'Vn', 'Enlnf', 'M3', 'En2lnf'});
elseif nargout>1
  [logS, En, Elnf, Vn, Enlnf] = S_terms(theta, nu, ...
      {'logS', 'En', 'Elnf', 'Vn', 'Enlnf'});
else
  [logS, En, Elnf] = S_terms(theta, nu, {'logS', 'En', 'Elnf'});
end

%b = logS / nu;
db = -logS/nu + theta.*En - Elnf;
 
if (nargout>1)
  dbp = nu*(theta.*Vn - Enlnf + En.*Elnf);
  
  if (nargout>2)
    ClnfV = En2lnf - 2*Enlnf.*En + Elnf.*(En.^2) ...
	    - Elnf.*Vn;
    
    dVn = nu*(theta.*M3 - ClnfV);
    dbpp = nu*(Vn+dVn);
    
    if 0
      % check derivative    
      delta = 0.000001;
      nu_del = exp(log(nu)+delta);
      [logS_del, En_del, Vn_del] = ...
	  S_terms(theta, nu_del, {'logS', 'En', 'Vn'});
      
      db_del  = (logS_del/nu_del - logS/nu) / delta;  % OKAY
      dbp_del = (En_del - En) / delta;                % OKAY
      dVn_del = (Vn_del - Vn) / delta;                % OKAY
      dbpp_del = (nu_del*Vn_del - nu*Vn) / delta;     % OKAY
      fprintf('%g %g %g %g\n', mean(abs([db_del - db]')), ...
	      mean(abs([dbp_del - dbp]')), ...
	      mean(abs([dVn_del - dVn]')), ...
	      mean(abs([dbpp_del - dbpp]')));
    %  keyboard
    end
    
  end
end



%if (nargout>1)
%  % NOT COMPLETELY TESTED [db2]
%  [logS, En, Vn, M3, Elnf, Elnf2, Enlnf] =  S_terms(theta, nu, [1 1 1 0 1 1 1]);
%  E1 = (theta.^2).*(Vn+En.*En) - 2*theta.*Enlnf + Elnf2;
%  E2 = theta.*En - Elnf;
%  
%  db2 = -db + nu*(E1 - E2.^2);
%end

% check with old
%olderr = sum(abs(db - db_hyp_OLD(theta, nu)));
%if (olderr > 1e-6)
%  fprintf('dbhyp-olderr=%g; ', olderr);
%end



%%% SPECIAL FUNCTIONS %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% d = 0 --> b
% d = 1 --> db
% d = 2 --> d2b
function out = b_combo_OLD(theta, nu, d)
%if (size(theta,1) == 1) && (size(theta,2) > 1)
%  warning('theta not column!');
%end

% work on each column of theta
out = zeros(size(theta));
for i=1:size(theta,2)
  % do not recompute if theta column does not change
  if (i>1) && all(theta(:,i-1)==theta(:,i))
    out(:,i) = out(:,i-1);
    
  else
    if (d == 0)
      [logsumts] = S_terms_OLD(theta(:,i), nu);
    else
      [logsumts, nts] = S_terms_OLD(theta(:,i), nu);
      n = 0:size(nts,2)-1;
    end
  
    switch(d)
     case 0   % b
      out(:,i) = logsumts / nu;
      
     case 1   % db
      nn = repmat(n, [size(theta,1),1]);
      out(:,i) = sum(nn .* nts, 2);
      
     case 2   % d2b
      nn  = repmat(n, [size(theta,1),1]);
      nn2 = nn.^2;
      out(:,i) = nu*(sum(nn2.*nts,2) - sum(nn.*nts,2).^2);
      
     otherwise
      error('not implemented or invalid');
    end
  end
end


% compute terms of S(theta, nu)
%   theta is a column vector
%   logsumts = log(sum(ts));
%   nts is a 2d matrix [length(theta) x MAXK]
function [logsumts, nts] = S_terms_OLD(theta, nu)
if (size(theta,2) ~= 1)
  error('theta is not a column vector!');
end

MAXK  = 100;  % default MAX is 100
retry = 1;
while(retry)
  retry=0;
  n = 0:MAXK;
  lts = nu*(theta*n - repmat(logfactlut(n),[size(theta,1),1]));

  % compute log(sum(exp(lts))) in a safe way
  % log(sum(exp(lts))) = log( sum(exp(lts-max)exp(max)) )
  %                      log( sum(exp(lts-max)) ) + max)
  ltsmax  = max(lts,[],2);
  mltsmax = repmat(ltsmax, [1,MAXK+1]);
  logsumts = log(sum(exp(lts-mltsmax),2)) + ltsmax;
  
  % compute bound
  lepsk = nu*theta - nu*log(MAXK+1);
  epsk = exp(lepsk);
  logbound = (MAXK+1)*nu*theta - nu*logfactlut(MAXK+1) ...
      - log(1-epsk) - logsumts;

  % error bound to high, recompute with larger MAXK.
  if any(logbound(:) > -10)
    %warning('error bound too high');
    %keyboard;
    if (MAXK < 500)
      retry = 1;
      MAXK = 500;
      fprintf('MAXK=%d (nu=%g) bound=%g; ', MAXK, nu, ...
	      min(logbound(:)));
%      keyboard
    end
  end
end
  
% normalize ts in log-domain
if nargout>1
  nts = exp(lts - repmat(logsumts, 1, MAXK+1));
end

%  keyboard


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% derivative of b wrt hyperparameter 
function db = db_hyp_OLD(theta, nu)
[logsumts, nts] = S_terms_OLD(theta, nu);

b = logsumts / nu;

% n
n = 0:size(nts,2)-1;

% n!
lnf = repmat(logfactlut(n), [size(theta,1),1]);

db = -b + sum((theta*n - lnf).* nts, 2);
      
%keyboard


% NEW VERSION %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% depending on the flags in mode, it calculates
%   mode(1)=1 -- log S
%   mode(2)=1 -- mean = E[n]
%   mode(3)=1 -- variance = E[n^2] - (E[n])^2
%   mode(4)=1 -- E[n^3] - (E[n])^3
%   mode(5)=1 -- E[log n!]
%   mode(6)=1 -- E[(log n!)^2]
%   mode(7)=1 -- E[n log n!]
function [logS, En, Vn, M3, Elfn, Elfn2, Enlfn] = S_terms_OLD2(theta, nu, mode)
if (size(theta,2) ~= 1)
  error('theta is not a column vector!');
end

doS  = mode(1); doM  = mode(2);
doV  = mode(3); doM3 = mode(4);
doF  = mode(5); doF2 = mode(6); doNF = mode(7);

% variance and third moment require the mean
if (doV || doM3)
  doM = 1;
end

% compute normalized likelihoods for moments
if any(mode(2:end))
  doNorm = 1;
else
  doNorm = 0;
end

% initialize storage
logS = []; En = []; Vn = [];  M3 = [];
Elfn = [];
if doS,  logS  = zeros(size(theta)); end
if doM,  En    = zeros(size(theta)); end
if doV,  Vn    = zeros(size(theta)); end
if doM3, M3    = zeros(size(theta)); end
if doF,  Elfn  = zeros(size(theta)); end
if doF2, Elfn2 = zeros(size(theta)); end
if doNF, Enlfn = zeros(size(theta)); end


% chunks for running approximation
% 0-50    --> MAXK = 100
% 50-100  --> MAXK = 250
% 100-250 --> MAXK = 500
% 250-500 --> MAXK = 1000
% 500+    --> Gaussian approximation
bbounds  = [0 50 100 250 500 1000];
lbbounds = log(bbounds);
numbounds = 4;    % up to 250-500
iiextra = [];

for i=1:numbounds
  % find theta within this range
  ii = find( (lbbounds(i) <= theta) .* (theta <= lbbounds(i+1)) );
  
  % get previous cases pushed to bigger range
  ii = [ii;iiextra];
  
  %fprintf('b%d=%d ', i, length(ii));
  
  if ~isempty(ii)
    MAXK = bbounds(i+2);
    
    % get log terms
    n = 0:MAXK;
    lts = nu*(theta(ii)*n - repmat(logfactlut(n),[length(ii),1]));

    % compute log(sum(exp(lts))) in a safe way
    % log(sum(exp(lts))) = log( sum(exp(lts-max)exp(max)) )
    %                      log( sum(exp(lts-max)) ) + max)
    ltsmax  = max(lts,[],2);
    mltsmax = repmat(ltsmax, [1,MAXK+1]);
    logsumts = log(sum(exp(lts-mltsmax),2)) + ltsmax;
  
    % compute bound
    lepsk = nu*theta(ii) - nu*log(MAXK+1);
    epsk = exp(lepsk);
    logbound = (MAXK+1)*nu*theta(ii) - nu*logfactlut(MAXK+1) ...
	- log(1-epsk) - logsumts;
    
    % error bound to high, so push these cases to the next round
    iiextra = find(logbound > -10);
    %if ~isempty(iiextra)
    %  fprintf('MAXK=%d (nu=%g) bound=%g (pushing %d); ', MAXK, nu, ...
    %	  max(logbound(:)), length(iiextra));
    %end
    
    if doNorm
      % create normalized distribution
      nts = exp(lts - repmat(logsumts, 1, MAXK+1));
    end
    
    % get logS
    if doS
      logS(ii) = logsumts;
    end
    
    % compute mean
    if doM
      nn = repmat(n, [length(ii),1]);
      En(ii) = sum(nn .* nts, 2);
    end
    
    % compute variance
    if doV
      nn2 = repmat(n.*n, [length(ii),1]);
      Vn(ii) = sum(nn2.*nts, 2) - En(ii).^2;
    end
    
    % compute third moment E[n^3] - (E[n])^3
    if doM3
      nn3 = repmat(n.^3, [length(ii),1]);
      M3(ii) = sum(nn3.*nts,2) - En(ii).^3;
    end
    
    % compute E(log n!)
    if doF || doF2 || doNF
      lnf = repmat(logfactlut(n), [length(ii),1]);
      if doF
	Elfn(ii) = sum(lnf.*nts, 2);
      end
      if doF2
	Elfn2(ii) = sum(lnf.*lnf.*nts, 2);
      end
      if doNF
	Enlfn(ii) = sum(nn.*lnf.*nts, 2);
      end
    end
   
  end  % end if ~empty(ii)
end % end for i

% now handle cases with large mean...
ii = find(theta >= lbbounds(numbounds+1));
%fprintf('b+=%d;\n', length(ii));
if ~isempty(ii)
  % for large mean, the distribution is approximately Gaussian
  % N(mu, mu/nu).
  mymu = exp(theta(ii));
  if doS
    % match peaks of Gaussian and lts to get scaling
    % max(ts) / S = G(0,mu/nu)
    % logS = max(log(ts)) - G(0,mu/nu)
    mmu = floor(mymu);
    logS(ii) = nu*(mmu.*theta(ii) - gammaln(mmu)) + ...
	0.5*log(2*pi*mmu/nu);
  end
  if doM   % mean
    En(ii) = mymu;
  end
  if doV   % variance
    Vn(ii) = mymu / nu;
  end
  if doM3
    % E[n^3] - E[n]^3 = 3mu var
    M3(ii) = 3*mymu.*(mymu/nu);
  end
  if doF || doF2 || doNF    % E[log n!]
    %warning('TODO! approximate now!');
    % just pick the mean and std points, and average
    tmps = sqrt(mymu/nu);
    tmp = [mymu, mymu-tmps, mymu+tmps];
    if doF
      Elfn(ii) = mean(gammaln(tmp), 2);
    end
    if doF2
      Elfn2(ii) = mean(gammaln(tmp).^2, 2);
    end
    if doNF
      Enlfn(ii) = mean(tmp.*gammaln(tmp), 2);
    end
  end    
%  keyboard
end



%%% special functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% try to make these numerically safe by:
%  1) forcing tail values
%  2) exp(x)/(1+exp(x)) in the log-domain first


% safe version of log(log(1+exp(eta)))
function th = theta(eta)
th = log(log(1+exp(eta)));

% for large eta, log(log(1+exp(eta))) = log(eta)
ii = find(eta>100);
th(ii) = log(eta(ii));

% for small negative eta, log(log(1+exp(eta))) = eta
jj = find(eta<-20);
th(jj) = eta(jj);


function dth = dtheta(eta)
tmp = log(1+exp(eta));
dth = exp(eta - tmp)./tmp;
% check negative tail (limit is 1)
ii = find(eta<=-35);
dth(ii) = 1;


function d2th = d2theta(eta)
tmp = log(1+exp(eta));
tmp2 = tmp.^2;
t1 = -exp(2*eta - 2*tmp)./tmp2;
t2 =  exp(eta   - 2*tmp)./tmp;
d2th = t1 + t2;
% check negative tail (limit is 0)
ii = find(eta<=-35);
d2th(ii) = 0;

function d3th = d3theta(eta)
tmp = log(1+exp(eta));
tmp2 = tmp.^2;
tmp3 = tmp.^3;

t1 =  2*exp(3*eta - 3*tmp)./tmp3;
t2 =   -exp(2*eta - 3*tmp)./tmp;
t3 =    exp(eta   - 3*tmp)./tmp;
t4 = -3*exp(2*eta - 3*tmp)./tmp2;

d3th = t1+t2+t3+t4;
% check negative tail (limit is 0)
%ii = find(eta<=-35);
%d3th(ii) = 0;

function ycheck = checkY(y)
ycheck = max(floor(y), 0);  % round and truncate
if any(y ~= ycheck)
 % error('y contains non-counting numbers!');
end
