function lf = likExGamma(hyp)
% Gamma likelihood function
%  domain: y>=0 
%   y<0 is mapped to y=0
% 
%  p(y) = [(nu/mu)^nu / Gamma(nu)] * y^(nu-1) e^(-y*nu/mu)
%
%   nu = exp(hyp)
%   hyp = log(nu)
%
% mapping is:  mu = g^{-1}(eta) = e^eta
%           : theta(eta) = -e^-eta
%
% the shape parameter is the dispersion parameter. However, this
% causes some problems when y=0 is observed, since this requires
% that nu=1.
%  We can prevent this problem by either:
%    1) setting y=0 to y=some small constant,
%    2) setting nu=1 (assume exponential distribution)
% 
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% 2010-11-23 -- ABC - fixed nu<1 issue.

% return number of likelihood parameters
if nargin<1, lf = 1; return; end

%%% get dispersion parameter %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
nu = exp(hyp);

%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% move negative y to zero
lf.chky    = @(y) checkY(y);

%%% setup Gamma %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1/nu;
lf.da      = @() -1/nu;
lf.b       = @(theta) -log(-theta);
lf.db      = @(theta) -1./theta;
lf.d2b     = @(theta) 1./(theta.^2);
lf.d3b     = @(theta) -2./(theta.^3);
lf.lh      = @(y) logh(y, nu);    % special function to handle y=0
lf.dlh     = @(y) dlogh(y, nu);
lf.theta   = @(eta) -exp(-eta);
lf.dtheta  = @(eta) exp(-eta);
lf.d2theta = @(eta) -exp(-eta);
lf.d3theta = @(eta) exp(-eta);
lf.fix_phi = 0;


%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c = 0.001;
lf.etahat  = @(y) log(y+c);
%lf.etahat  = @(y) log(y);

%%% special functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function lh = logh(y, nu)

lh = (nu-1)*log(y) + nu*log(nu) - gammaln(nu);

% fix zero entries
ii = find(y==0);
if ~isempty(ii)
  if (nu>1)
    lh(ii) = -inf;  % y^(nu-1) = 0;
  elseif (nu==1)
    lh(ii) = 0;    % 0 log 0 + 1 log 1 - gammaln(1) = 0 + 0 - 0 = 0
  else  % 0<nu<1   % (nu-1)*log 0 + ... = inf
    lh(ii) = inf;
    %warning('y==0: perhaps this is not the best model!')
  end
end

function dlh = dlogh(y, nu)
dlh = nu*(log(y) + log(nu) + 1 - psi(nu));

% fix zero entries
ii = find(y==0);
if ~isempty(ii)
  if (nu>0)
    dlh(ii) = -inf;  % y^nu = 0
  elseif (nu==0)
    dlh(ii) = 0;      % 0 (log 0 + log 0 + 1 - log 0) = 0log0 = 0
  else
    error('TODO - dlogh for nu<0!');
  end
end

function y = checkY(y)
 ii = find(y<0);
 if ~isempty(ii)
   y(ii) = 0;
   warning('y has negative values!');
 end