function lf = likExGammaScale(hyp)
% Gamma likelihood function w/ scale hyperparameter
%  domain: y>=0 
%   y<0 is mapped to y=0
% 
% shape is tied to the scale parameter, and the scale parameter is
% now the dispersion
%
%   a(phi)    = scale
%   T(y)      = log(y)
%  theta(eta) = mean 
%  b(theta,phi) = theta log(phi) + phi log Gamma(theta/phi)
%
%  hyp = log(scale)
%
%  note: mean = scale*shape
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% formerly called likExGammaShapeA.m

% return number of likelihood parameters
if nargin<1, lf = 1; return; end


%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% move negative y to zero
lf.chky    = @(y) checkY(y);

% get dispersion
phi = exp(hyp(1));
logphi = hyp(1);

%%% setup Gamma %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() phi;
lf.da      = @() phi;

lf.b       = @(theta) b_combo(theta, phi, 0);   % capture hyperparameter
lf.db      = @(theta) b_combo(theta, phi, 1);
lf.d2b     = @(theta) b_combo(theta, phi, 2);
lf.d3b     = @(theta) b_combo(theta, phi, 3);
lf.b_and_d = @(theta) b_and_d(theta, phi);

lf.lh      = @(y) logh(y, phi);
lf.dlh     = @(y) dlogh(y, phi);

lf.theta   = @(eta) exp(eta);
lf.dtheta  = @(eta) exp(eta);
lf.d2theta = @(eta) exp(eta);
lf.d3theta = @(eta) exp(eta);
lf.fix_phi = 0;

% approximate y=0 with 1e-6
%lf.T       = @(y) log(y.*(y>1e-6) + (1e-6)*(y<=1e-6));
lf.T       = @(y) log(y);
lf.Tinv    = @(x) exp(x);


% special case: b has hyperparameter
lf.b_has_hyp  = 1;
lf.db_hyp     = @(theta) db_hyp(theta, phi);


%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%c = 0.001;
%lf.etahat  = @(y) log(y+c);

% the ideal expansion point (as a function of loghyp)
lf.etahat   = @(y) etahat(y, logphi);
lf.etahat_has_hyp = 1;
lf.detahat_hyp    = @(y) detahat_hyp(y, logphi);

%lf.check = @check;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [bout] = b_combo(theta, phi, d)
switch(d)
 case 0
  bout = theta*log(phi) + phi*gammaln(theta/phi);
 case 1
  bout = log(phi) + psi(0,theta/phi);
 case 2
  bout = psi(1,theta/phi)/phi;
 case 3
  bout = psi(2,theta/phi)/(phi^2);
 otherwise
  error('bad d');
end

function [b, db, d2b, d3b] = b_and_d(theta, phi)

if (nargout==2)
  tmp = psi(0, theta/phi);
  psi0 = reshape(tmp, size(theta));
elseif (nargout==3)
  tmp = psi(0:1, theta/phi);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
elseif (nargout==4)
  tmp = psi(0:2, theta/phi);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
  psi2 = reshape(tmp(3,:), size(theta));
end

b = theta*log(phi) + phi*gammaln(theta/phi);
if nargout>1
  db = log(phi) + psi0;
  if (nargout>2)
    d2b = psi1/phi;
    if (nargout>3)
      d3b = psi2/(phi^2);
    end
  end
end


function [db, dbp, dbpp] = db_hyp(theta, phi)

if (nargout==1)
  tmp = psi(0, theta/phi);
  psi0 = reshape(tmp, size(theta));
elseif (nargout==2)
  tmp = psi(0:1, theta/phi);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
elseif (nargout==3)
  tmp = psi(0:2, theta/phi);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
  psi2 = reshape(tmp(3,:), size(theta));
end

%db = theta/phi + gammaln(theta/phi) - (theta/phi .* psi0);
db = theta + phi*gammaln(theta/phi) - (theta .* psi0);
if (nargout>1)
  %dbp = 1/phi - (theta/(phi^2)).*psi1;
  dbp = 1 - (theta/phi).*psi1;
  if (nargout>2)
    %dbpp = -psi1/(phi^2) - theta.*psi2 / (phi^3);
    dbpp = -psi1/phi - theta.*psi2 / (phi^2);
  end
end


function lh = logh(y, phi)
y(find(y<=1e-6)) = 1e-6;
lh = -log(y) - y/phi;

% handle special case: y=0
if 0
  ii = find(y==0);
  if ~isempty(ii)
    % p(y=0) = y^(k-1) e^(-y/scale) / (scale^k gamma(k))
    %        = 0 * 1 = 0  (when k>1)
    % p(y=0) = inf, when 0<k<1
    y0 = 1e-6;
    lh(ii) = -log(y0) - y0/phi;
  end
end

function dlh = dlogh(y, phi)
y(find(y<=1e-6)) = 1e-6;
dlh = y/phi;


function deta = detahat_hyp(y, logphi)
ipsi = invpsi(log(y) - logphi);
deta = 1 - 1./(ipsi.*psi(1, ipsi));

function eh = etahat(y, logphi)
eh = logphi+log(invpsi(log(y)-logphi));

function y = checkY(y)
ii = find(y<0);
if ~isempty(ii)
  y(ii) = 0;
  warning('y has negative values!');
end
