function lf = likExGammaScale1(hyp)
% Gamma likelihood function w scale hyp=1
%  domain: y>=0 
%   y<0 is mapped to y=0 
% 
% shape parameter varies, and scale parameter is fixed to 1
%
%  p(y) = e^(-y) y^(k-1) / Gamma(k)
%
%  log p(y) = -y + (k-1)*log y - log Gamma(k)
%
%  Hence, T(y)     = log(y),
%         theta    = k 
%         b(theta) = log Gamma(k)
%         h(y)     = -y - log(y)
%
%  Mapping: theta = exp(eta) > 0 
%          i.e., k>0
%
% note k = nu
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% formerly called likExGammaShapeA.m

% return number of likelihood parameters
if nargin<1, lf = 0; return; end

%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% move negative y to zero
lf.chky    = @(y) checkY(y);

% force mu close to zero to be zero?

%%% setup Gamma %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1;
lf.da      = @() 0;
lf.b       = @(theta) gammaln(theta);
lf.db      = @(theta) psi(0,theta);
lf.d2b     = @(theta) psi(1,theta);
lf.d3b     = @(theta) psi(2,theta);
lf.b_and_d = @(theta) b_and_d(theta);
lf.lh      = @(y) -y-log(y);
lf.dlh     = @(y) 0;
lf.theta   = @(eta) exp(eta);
lf.dtheta  = @(eta) exp(eta);
lf.d2theta = @(eta) exp(eta);
lf.d3theta = @(eta) exp(eta);
lf.fix_phi = 1;

lf.T       = @(y) log(y);
lf.Tinv    = @(x) exp(x);

%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%c = 0.001;
%lf.etahat  = @(y) log(y+c);
%lf.etahat = @(y) log(y);

% the ideal expansion point (much better than above)!
lf.etahat = @(y) log(invpsi(log(y)));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [b, db, d2b, d3b] = b_and_d(theta)

if (nargout==2)
  tmp = psi(0, theta);
  psi0 = reshape(tmp, size(theta));
elseif (nargout==3)
  tmp = psi(0:1, theta);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
elseif (nargout==4)
  tmp = psi(0:2, theta);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
  psi2 = reshape(tmp(3,:), size(theta));
end

b = gammaln(theta);
if nargout>1
  db = psi0;
  if (nargout>2)
    d2b = psi1;
    if (nargout>3)
      d3b = psi2;
    end
  end
end

function y = checkY(y)
ii = find(y<0);
if ~isempty(ii)
  y(ii) = 0;
  warning('y has negative values!');
end
