function lf = likExGammaScale2(hyp)
% Gamma likelihood function
%  domain: y>=0 
%   y<0 is mapped to y=0 
% 
% shape parameter varies, and scale parameter is fixed to 2
% in this case, Gamma distribution reduced to chi-squared distribution
%
%  p(y|k) = e^(-y/2) y^(k/2-1) / Gamma(k/2)/2^(k/2)
%
%  log p(y|k) = -y/2 - log y + (k / 2)*log y - log Gamma(k/2) - (log 2/2)*k
%
%  Hence, T(y)     = log(y),
%         theta    = k 
%         b(theta) = k*log 2 + 2*log Gamma(k/2)   
%         log h(y) = -y/2 - log(y)
%         phi      = 2
%  Mapping: theta = exp(eta) > 0 
%          i.e., k>0
%
% note k = nu
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27

% return number of likelihood parameters
if nargin<1, lf = 0; return; end

%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% move negative y to zero
lf.chky    = @(y) checkY(y);

% force mu close to zero to be zero?

%%% setup Gamma %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 2;
lf.da      = @() 0;
lf.b       = @(theta) theta * log(2) + 2 * gammaln(theta / 2);
lf.db      = @(theta) log(2) + psi(0,theta / 2);
lf.d2b     = @(theta) psi(1,theta / 2) / 2;
lf.d3b     = @(theta) psi(2,theta / 2) / 4;
lf.b_and_d = @(theta) b_and_d(theta);
lf.lh      = @(y) -y/2-log(y);
lf.dlh     = @(y) 0;
lf.theta   = @(eta) exp(eta);
lf.dtheta  = @(eta) exp(eta);
lf.d2theta = @(eta) exp(eta);
lf.d3theta = @(eta) exp(eta);
lf.fix_phi = 1;

lf.T       = @(y) log(y);
lf.Tinv    = @(x) exp(x);

%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%c = 0.001;
%lf.etahat  = @(y) log(y+c);
%lf.etahat = @(y) log(y);

% the ideal expansion point (much better than above)!
lf.etahat = @(y) log(2*invpsi(log(y)-log(2)));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

function [b, db, d2b, d3b] = b_and_d(theta)

if (nargout==2)
  tmp = psi(0, theta/2);
  psi0 = reshape(tmp, size(theta));
elseif (nargout==3)
  tmp = psi(0:1, theta/2);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
elseif (nargout==4)
  tmp = psi(0:2, theta/2);
  psi0 = reshape(tmp(1,:), size(theta));
  psi1 = reshape(tmp(2,:), size(theta));
  psi2 = reshape(tmp(3,:), size(theta));
end

b = theta * log(2) + 2 * gammaln(theta / 2);
if nargout>1
  db = log(2) + psi0;
  if (nargout>2)
    d2b = psi1 / 2;
    if (nargout>3)
      d3b = psi2 / 4;
    end
  end
end

function y = checkY(y)
ii = find(y<0);
if ~isempty(ii)
  y(ii) = 0;
  warning('y has negative values!');
end
