function lf = likExInvGauss(hyp)
% Inverse Gaussian likelihood function
%  domain: y>=0 
%   y<0 is mapped to y=0
% 
%  p(y) = (lambda/(2pi y^3))^{0.5} e^{-lambda (y-mu)^2 / (2 mu^2 y)}
%
%   hyp = log(lambda)
%  theta = -1/(2 mu^2)
%
% mapping is:  mu = g^{-1}(eta) = e^(eta/2) / sqrt(2)
%           : theta(eta) = -e^-eta
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27


% return number of likelihood parameters
if nargin<1, lf = 1; return; end


%%% get dispersion parameter %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
loglambda = hyp;
lambda = exp(hyp);

%%% check/process y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% move negative y to zero
lf.chky    = @(y) checkY(y);


%%% setup InvGauss %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1/lambda;
lf.da      = @() -1/lambda;
lf.b       = @(theta) -sqrt(-2*theta);
lf.db      = @lf_db;
lf.d2b     = @lf_d2b;
lf.d3b     = @lf_d3b;
lf.lh      = @(y) logh(y,loglambda);
lf.dlh     = @(y) 0.5-0.5*lambda./y;
lf.theta   = @(eta) -exp(-eta);
lf.dtheta  = @(eta) exp(-eta);
lf.d2theta = @(eta) -exp(-eta);
lf.d3theta = @(eta) exp(-eta);
lf.fix_phi = 0;


%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%c = 0.001;
lf.etahat = @(y) 2*log(y)+log(2);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function lh = logh(y, loglambda)

lh = -0.5*log(2*pi)-1.5*log(y)+0.5*loglambda-0.5*exp(loglambda)./y;

% set p(y=0) = 0
ii = find(y==0);
if ~isempty(ii)
  lh(ii) = -inf;
end


% special functions for handling derivatives of b
% make sure that (-0)^(-1) = inf
% theta is negative, so (-2*theta)^(-x) = inf
function db = lf_db(theta)
db = (-2*theta).^(-0.5);
db(theta > -1e-50) = inf;

function d2b = lf_d2b(theta)
d2b = (-2*theta).^(-1.5);
d2b(theta > -1e-50) = inf;

function d3b = lf_d3b(theta)
d3b = 3*(-2*theta).^(-2.5);
d3b(theta > -1e-50) = inf;

function y = checkY(y)
ii = find(y<0);
if ~isempty(ii)
  y(ii) = 0;
  warning('y has negative values!');
end