function lf = likExPoissonLinear(hyp)
% Poisson likelihood function with linear-mapped mean
%   y in {0,1,2,3,...}
%
% eta is mapped to a nearly linear mean using the logistic-loss function
%
%   mean = g^{-1}(eta) = log(1+e^eta)
%
%  theta(eta) = log(log(1+e^eta))
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27



% return number of likelihood parameters
if nargin<1, lf = 0; return; end

%%% check y %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.chky    = @(y) checkY(y);

%%% setup Gaussian %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lf.a       = @() 1;
lf.da      = @() 0;
lf.b       = @(theta) exp(theta);
lf.db      = @(theta) exp(theta);
lf.d2b     = @(theta) exp(theta);
lf.d3b     = @(theta) exp(theta);
lf.lh      = @(y) -logfactlut(y);  %gammaln(y+1);
lf.dlh     = @(y) zeros(size(y));
lf.theta   = @(eta) theta(eta);
lf.dtheta  = @dtheta;
lf.d2theta = @d2theta;
lf.d3theta = @d3theta;
lf.fix_phi = 1;         % no dispersion parameter


%%% Taylor approximation %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
c = 0.0001;
lf.etahat  = @(y) log(exp(y)-1+c);

%%% special functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% try to make these numerically safe by:
%  1) forcing tail values
%  2) exp(x)/(1+exp(x)) in the log-domain first


% safe version of log(log(1+exp(eta)))
function th = theta(eta)
th = log(log(1+exp(eta)));

% for large eta, log(log(1+exp(eta))) = log(eta)
ii = find(eta>100);
th(ii) = log(eta(ii));

% for small negative eta, log(log(1+exp(eta))) = eta
jj = find(eta<-20);
th(jj) = eta(jj);


function dth = dtheta(eta)
tmp = log(1+exp(eta));
dth = exp(eta - tmp)./tmp;
% check negative tail (limit is 1)
ii = find(eta<=-35);
dth(ii) = 1;


function d2th = d2theta(eta)
tmp = log(1+exp(eta));
tmp2 = tmp.^2;
t1 = -exp(2*eta - 2*tmp)./tmp2;
t2 =  exp(eta   - 2*tmp)./tmp;
d2th = t1 + t2;
% check negative tail (limit is 0)
ii = find(eta<=-35);
d2th(ii) = 0;

function d3th = d3theta(eta)
% % check negative tail (limit is 0)
ii = find(eta<=-35);
tmp = log(1+exp(eta));
tmp(ii) = 1.0000e-010;
tmp2 = tmp.^2;
tmp3 = tmp.^3;

t1 =  2*exp(3*eta - 3*tmp)./tmp3;
t2 =   -exp(2*eta - 3*tmp)./tmp;
t3 =    exp(eta   - 3*tmp)./tmp;
t4 = -3*exp(2*eta - 3*tmp)./tmp2;

d3th = t1+t2+t3+t4;

function ycheck = checkY(y)
ycheck = max(floor(y), 0);  % round and truncate
if any(y ~= ycheck)
%  error('y contains non-counting numbers!');
end

