function [thall] = theta_linlog()
% linear-log theta function
%
%   theta(eta) = log(log(1+e^eta))
%
% If the canonical link has mu = g^{-1}(eta) = exp(eta), this theta
% has the property to map mu = g^{-1}(eta) = log(1+e^eta), which is
% the logistic-loss function.
% i.e., the mean is a linear function of eta when eta>0, and
%        approaches zero when eta<0
%
% Generalized Gaussian Process Models Toolbox (GPML Add-on)
% Copyright (c) by Antoni B. Chan, 2013-11-27


thall.th   = @th;
thall.dth  = @dth;
thall.d2th = @d2th;
thall.d3th = @d3th;


%%% special functions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% try to make these numerically safe by:
%  1) forcing tail values
%  2) exp(x)/(1+exp(x)) in the log-domain first


% safe version of log(log(1+exp(eta)))
function th = theta(eta)
th = log(log(1+exp(eta)));

% for large eta, log(log(1+exp(eta))) = log(eta)
ii = find(eta>100);
th(ii) = log(eta(ii));

% for small negative eta, log(log(1+exp(eta))) = eta
jj = find(eta<-20);
th(jj) = eta(jj);


function dth = dtheta(eta)
tmp = log(1+exp(eta));
dth = exp(eta - tmp)./tmp;
% check negative tail (limit is 1)
ii = find(eta<=-35);
dth(ii) = 1;


function d2th = d2theta(eta)
tmp = log(1+exp(eta));
tmp2 = tmp.^2;
t1 = -exp(2*eta - 2*tmp)./tmp2;
t2 =  exp(eta   - 2*tmp)./tmp;
d2th = t1 + t2;
% check negative tail (limit is 0)
ii = find(eta<=-35);
d2th(ii) = 0;


function d3th = d3theta(eta)
tmp = log(1+exp(eta));
tmp2 = tmp.^2;
tmp3 = tmp.^3;

t1 =  2*exp(3*eta - 3*tmp)./tmp3;
t2 =   -exp(2*eta - 3*tmp)./tmp;
t3 =    exp(eta   - 3*tmp)./tmp;
t4 = -3*exp(2*eta - 3*tmp)./tmp2;

d3th = t1+t2+t3+t4;
% check negative tail (limit is 0)
%ii = find(eta<=-35);
%d3th(ii) = 0;



